/*
 * linux/video/encode_sjis.c
 *
 * Copyright (C) 2000		Kazuhide Takahashi
 *				HOLON Inc.
 * Copyright (C) 1999		Christopher Li, Jim Chen
 *				GNU/Linux Research Center
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 *
 *
 */


#include <linux/module.h>
#include <linux/fb_doublebyte.h>
#include "font_jis16.h"
#define min1 0x81
#define max1 0xfc


int is_left(int c)
{
	return ( (c >= 0x81 && c <= 0x9f) || (c >= 0xe0 && c <= 0xfc) );
}
int is_right(int c)
{
	return ( (c >= 0x40 && c <= 0x7e) || (c >= 0x80 && c <= 0xfc) );
}
int index_sjis (int left, int right)
{
	int ch1, ch2;

	if ( !is_left(left) || !is_right(right) ) return -512;

	ch1 = left;
	ch2 = right;

	if ( ch1 <= 0x9f )
		ch1 -= 0x71;
	else
		ch1 -= 0xb1;
	ch1 *= 2;
	++ch1;

	if ( ch2 >= 0x7f )
		--ch2;
	if ( ch2 >= 0x9e ) {
		ch2 -= 0x7d;
		++ch1;
	} else {
		ch2 -= 0x1f;
	}

	if (ch1 > 0x29)
		return ((ch2 - 0x40 + (ch1 - 0x25) * 96) << 5);
	else
		return ((ch2 - 0x20 + (ch1 - 0x20) * 96) << 5);
}
struct double_byte db_gb =
{
	0,
	"SJIS",
	is_left,
	is_right,
	index_sjis,
	16,16,
	max_jis16,
	font_jis16
};

int init_module(void)
{
	if (doublebyte_default) return 1;
	doublebyte_default = &db_gb;
	return 0;
}

void cleanup_module(void)
{
	doublebyte_default = (void*) 0;
}	
