#!/bin/sh
# Simple stupid automated kernel build script
# Copyright 2007 by Andre Beckedorf < evilJazz AT katastrophos DOT net >
# This script is released under the terms of the GNU Public Licencse version 2 or higher.

function ApplySpecificPatches_pdaXrom
{
	patch -f -s -p1 $1 < ../specific-patches/pdaxrom/ide-cs.c.diff
}

function ApplySpecificConfig_pdaXrom
{
	sed \
	-e 's/CONFIG_ARCH_SHARP_SL_E=y/# CONFIG_ARCH_SHARP_SL_E is not set/g' \
	-e 's/# CONFIG_ARCH_SHARP_SL_J is not set/CONFIG_ARCH_SHARP_SL_J=y/g' \
	-e 's/CONFIG_PPP_DEFLATE=y/CONFIG_PPP_DEFLATE=m/g' \
	-e 's/# CONFIG_AIRO_CS is not set/CONFIG_AIRO_CS=m/g' \
	-e 's/CONFIG_SCSI=m/CONFIG_SCSI=y/g' \
	-e 's/CONFIG_BLK_DEV_SD=m/CONFIG_BLK_DEV_SD=y/g' \
	-e 's/CONFIG_BLK_DEV_SR=m/CONFIG_BLK_DEV_SR=y/g' \
	-e 's/CONFIG_CHR_DEV_SG=m/CONFIG_CHR_DEV_SG=y/g' \
	-e 's/CONFIG_SCSI_MULTI_LUN=y/# CONFIG_SCSI_MULTI_LUN is not set/g' \
	-e 's/CONFIG_INPUT_KEYBDEV=y/CONFIG_INPUT_KEYBDEV=m/g' \
	-e 's/CONFIG_INPUT_MOUSEDEV=y/CONFIG_INPUT_MOUSEDEV=m/g' \
	-e 's/CONFIG_INPUT_EVDEV=m/CONFIG_INPUT_EVDEV=y/g' \
	-e 's/CONFIG_BOOT_PRESSURE_ON=y/# CONFIG_BOOT_PRESSURE_ON is not set/g' \
	-e 's/CONFIG_PCMCIA_SERIAL_CS=m/CONFIG_PCMCIA_SERIAL_CS=y\nCONFIG_PCMCIA_CHRDEV=y/g' \
	-e 's/CONFIG_SQUASHFS=y/CONFIG_SQUASHFS=m/g' \
	-e 's/CONFIG_ISO9660_FS=m/CONFIG_ISO9660_FS=y/g' \
	-e 's/# CONFIG_NTFS_RW is not set/CONFIG_NTFS_RW=y/g' \
	-e 's/CONFIG_ZISOFS_FS=m/CONFIG_ZISOFS_FS=y/g' \
	-e 's/CONFIG_NLS_DEFAULT="iso8859-1"/CONFIG_NLS_DEFAULT="utf8"/g' \
	-e 's/CONFIG_NLS_CODEPAGE_866=y/CONFIG_NLS_CODEPAGE_866=m/g' \
	-e 's/CONFIG_NLS_CODEPAGE_932=y/CONFIG_NLS_CODEPAGE_932=m/g' \
	-e 's/CONFIG_NLS_CODEPAGE_1251=y/CONFIG_NLS_CODEPAGE_1251=m/g' \
	-e 's/CONFIG_NLS_ISO8859_5=y/CONFIG_NLS_ISO8859_5=m/g' \
	-e 's/CONFIG_NLS_KOI8_R=y/CONFIG_NLS_KOI8_R=m/g' \
	-e 's/CONFIG_NLS_KOI8_U=y/CONFIG_NLS_KOI8_U=m/g' \
	-e 's/CONFIG_UNICON=y/# CONFIG_UNICON is not set/g' \
	-e 's/CONFIG_UNICON_GB=m//g' \
	-e 's/CONFIG_UNICON_GB=m//g' \
	-e 's/CONFIG_UNICON_SJIS=m//g' \
	-e 's/CONFIG_UNICON_SJIS=m//g' \
	-e 's/CONFIG_UNICON_EUCJP=m//g' \
	-e 's/CONFIG_UNICON_EUCJP=m//g' \
	-e 's/# CONFIG_USB_DEBUG is not set/CONFIG_USB_DEBUG=y/g' \
	-e 's/# CONFIG_MMC is not set/CONFIG_MMC=m\nCONFIG_MMC_PXA=m\n# CONFIG_MMC_PXA_PIO is not set\nCONFIG_MMC_PXA_CRC_BUG=y\nCONFIG_MMC_BLOCK=m\nCONFIG_MMC_PARTITIONS=y/g' \
	-e 's/# CONFIG_BLUEZ_HCIUART_BCSP_TXCRC is not set/CONFIG_BLUEZ_HCIUART_BCSP_TXCRC=y/g' \
	.config > .newconfig
	mv .newconfig .config
}

function ApplyNWFPE
{
	sed -e 's/# CONFIG_FPE_NWFPE is not set/CONFIG_FPE_NWFPE=y/g' \
	-e 's/CONFIG_FPE_FASTFPE=y/# CONFIG_FPE_FASTFPE is not set/g' .config > .newconfig
	mv .newconfig .config
}

function ApplyFastFPE
{
	sed -e 's/# CONFIG_FPE_FASTFPE is not set/CONFIG_FPE_FASTFPE=y/g' \
	-e 's/CONFIG_FPE_NWFPE=y/# CONFIG_FPE_NWFPE is not set/g' .config > .newconfig
	mv .newconfig .config
}

function ApplyCachedFB
{
	sed -e 's/# CONFIG_SHARPSL_PXA_CONSISTENT_ALLOC is not set/CONFIG_SHARPSL_PXA_CONSISTENT_ALLOC=y/g' .config > .newconfig
	mv .newconfig .config
}

function ApplyUncachedFB
{
	sed -e 's/CONFIG_SHARPSL_PXA_CONSISTENT_ALLOC=y/# CONFIG_SHARPSL_PXA_CONSISTENT_ALLOC is not set/g' .config > .newconfig
	mv .newconfig .config
}

function MakeZImage
{
	echo "    Building kernel $2"

	if [ "$VERBOSE" -eq "1" ]; then
		echo | make oldconfig
		make dep PREPROC="$PREPROC"
		make clean
		make -j4 zImage PREPROC="$PREPROC"
	else
		echo | make oldconfig > /dev/null 2>&1
		make dep PREPROC="$PREPROC" > /dev/null 2>&1
		make clean > /dev/null 2>&1
		make -j4 zImage PREPROC="$PREPROC" > /dev/null 2>&1
	fi
	
	mkdir -p "../images/$1/$2"
	cp arch/arm/boot/zImage "../images/$1/$2/"
	cp .config "../images/$1/$2/kernel_config"
	OLDPWD="$PWD"
	cd "../images/$1/$2/"
	md5sum "zImage" > "zImage.md5"
	cd "$OLDPWD"
}

function MakeModules
{
	echo "    Building modules $2"

	rm -Rf /tmp/modules
	mkdir -p /tmp/modules

	if [ "$VERBOSE" -eq "1" ]; then
		make -j4 modules PREPROC="$PREPROC"
		make modules_install INSTALL_MOD_PATH=/tmp/modules
	else
		make -j4 modules PREPROC="$PREPROC" > /dev/null 2>&1
		make modules_install INSTALL_MOD_PATH=/tmp/modules > /dev/null 2>&1
	fi
	
	mkdir -p "../images/$1"
	OLDPWD="$PWD"
	cd /tmp/modules
	tar czf "$OLDPWD/../images/$1/$2-modules.tar.gz" lib
	rm -Rf /tmp/modules
	cd "$OLDPWD"
	cd "../images/$1/"
	md5sum "$2-modules.tar.gz" > "$2-modules.tar.gz.md5"
	cd "$OLDPWD"
}

function BuildConfig_SharpROM
{
	echo "  Building config $1"

	cd linux
	make mrproper > /dev/null 2>&1
	make distclean > /dev/null 2>&1

	cp "../configs/$1" .config
	ApplyCachedFB
	ApplyFastFPE
	MakeZImage "$1" "sharprom-kernel-fastfpe-cachedfb"

	ApplyNWFPE
	MakeZImage "$1" "sharprom-kernel-nwfpe-cachedfb"

	ApplyUncachedFB
	MakeZImage "$1" "sharprom-kernel-nwfpe-uncachedfb"

	ApplyFastFPE
	MakeZImage "$1" "sharprom-kernel-fastfpe-uncachedfb"
	
	MakeModules "$1" "sharprom"
	
	cd ..
}

function BuildConfig_pdaXrom
{
	echo "Building $1:"

	cd linux
	make mrproper > /dev/null 2>&1
	make distclean > /dev/null 2>&1

	cp "../configs/$1" .config
	ApplySpecificPatches_pdaXrom
	ApplySpecificConfig_pdaXrom
	ApplyCachedFB
	ApplyFastFPE
	MakeZImage "$1" "pdaxrom-kernel-fastfpe"

	ApplyNWFPE
	MakeZImage "$1" "pdaxrom-kernel-nwfpe"

	MakeModules "$1" "pdaxrom"
	
	ApplySpecificPatches_pdaXrom -R > /dev/null 2>&1
	
	cd ..
}

# If you want to see all messages during compilation,
# set the following to 1 for verbosity.
VERBOSE=0

# Determine if and where ccache is installed...
CCACHE=`which ccache`

if [ "$CCACHE" != "" ]
then
    echo "Using ccache."
    PREPROC="$CCACHE"    
else
    echo "Not using ccache."
    PREPROC=""
fi

echo
echo "Building SharpROM kernels"
BuildConfig_SharpROM SL-C3200-terrier
BuildConfig_SharpROM SL-C3100-borzoi
BuildConfig_SharpROM SL-C3000-spitz
BuildConfig_SharpROM SL-C1000-akita

echo
echo "Building pdaXrom kernels"
BuildConfig_pdaXrom SL-C3200-terrier
BuildConfig_pdaXrom SL-C3100-borzoi
BuildConfig_pdaXrom SL-C3000-spitz
BuildConfig_pdaXrom SL-C1000-akita
